# Python interface to Basepair's REST API
#
# Copyright [2013] - [2017] Basepair LLC
# All Rights Reserved.
#
# NOTICE: All information contained herein is, and remains
# the property of Basepair LLC and its suppliers, if any.
# The intellectual and technical concepts contained
# herein are proprietary to Basepair LLC
# and its suppliers and may be covered by U.S. and Foreign Patents,
# patents in process, and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material
# is strictly forbidden unless prior written permission is obtained
# from Basepair LLC.

import os
import sys
import json
import subprocess
from subprocess import CalledProcessError
import time
import datetime

import requests
from requests.exceptions import ConnectionError

requests.packages.urllib3.disable_warnings()


class BpInterface():
    """ A wrapper over the REST API for accessing the Basepair system

    Use it thus:

    > import basepair
    > bp = basepair.connect(conf)

    conf is a JSON object with at least these keys:
    {
        "api": {
            "host": "http://app.basepairtech.com/",
            "prefix": "api/v1/",
            "username": "utk",
            "api_key": "xxx"
        },
        "aws": {
            "s3": {
               "bucket": "basepair"
            }
        }
    }

    Some of the common methods are:

    # create a sample
    sample_id = bp.create_sample(info_dict)

    #  create an analysis
    analysis_id = bp.create_analysis(workflow_id, sample_id)

    # download results - bam, dedup file
    filename = bp.get_file_by_tags(sample_id, ['bam', 'dedup'])

    # delete sample
    bp.delete_sample(sample_id)
    """

    def __init__(self, conf, scratch=None, use_cache=False, verbose=False):
        self.verbose = verbose
        self.conf = conf
        if 'AWS_CONFIG_FILE' not in os.environ:
            os.environ['AWS_ACCESS_KEY_ID'] = self.conf['aws']['aws_id']
            os.environ['AWS_SECRET_ACCESS_KEY'] = \
                self.conf['aws']['aws_secret']

        if not scratch:
            scratch = '.'
        self.scratch = conf.get('scratch', scratch).rstrip('/')
        self.use_cache = use_cache
        if use_cache and self.verbose:
            print >>sys.stderr, 'Warning: caching info'

        self.payload = {
            'username': self.conf['api']['username'],
            'api_key': self.conf['api']['api_key']
        }
        self.headers = {'content-type': 'application/json'}
        self.get_userid()

    def get_request(self, url, user_params=None, verify=False):
        """Add params to GET request"""
        params = self.payload
        if user_params:
            params.update(user_params)

        try:
            res = requests.get(url, params=params, verify=verify)
        except ConnectionError:
            print >>sys.stderr, 'connection error'
            print >>sys.stderr, url
            return None, None

        data = res.json() if res.status_code == 200 else None

        if res.status_code != 200:
            print >>sys.stderr, 'code:', res.status_code, 'for', url

        return data, res.status_code

    def post_request(self, url, data, verify=False):
        """Add params to POST request"""
        return requests.post(url, data=json.dumps(data), params=self.payload,
                             headers=self.headers, verify=verify)

    def put_request(self, url, data, verify=False):
        """Add params to PUT request, data is pure json"""
        return requests.put(url, data=json.dumps(data), params=self.payload,
                            headers=self.headers, verify=verify)

    def patch_request(self, url, data, verify=False):
        """Add params to PATCH request"""
        return requests.patch(url, data=json.dumps(data), params=self.payload,
                              headers=self.headers, verify=verify)

    def delete_request(self, url, verify=False):
        """Add params to DELETE request"""
        return requests.delete(url, params=self.payload, verify=verify)

    def get_url(self, kind=None, uid=None):
        """Create URL for a given object kind, add id if provided"""
        # url = self.conf['api']['url'] + kind
        url = self.conf['api']['host'] + self.conf['api']['prefix'] + kind
        if uid:
            url = '{}/{}'.format(url, uid)

        return url

    def get_user_url(self, uid=None):
        """Get URL for accessing 1 sample or all samples"""
        return self.get_url('users', uid)

    def get_sample_url(self, uid=None):
        """Get URL for accessing 1 sample or all samples"""
        return self.get_url('samples', uid)

    def get_project_url(self, uid=None):
        """Get URL for accessing 1 project or all projects"""
        return self.get_url('projects', uid)

    def get_group_url(self, uid=None):
        """Get URL for accessing 1 group or all groups"""
        return self.get_url('groups', uid)

    def get_analysis_url(self, uid=None):
        """Get URL for accessing 1 analysis or all analysis"""
        return self.get_url('analyses', uid)

    def get_file_url(self, uid=None):
        """Get URL for accessing 1 file or all files"""
        return self.get_url('files', uid)

    def get_upload_url(self, uid=None):
        """Get URL for accessing 1 upload or all uploads"""
        return self.get_url('uploads', uid)

    def parse_url(self, url):
        """Parse URL to get the id, and other stuff"""
        parts = url.rsplit('/', 1)
        return {
            'id': parts[1]
        }

    def get_id_from_url(self, url):
        """Parse URL to get the id"""
        return self.parse_url(url)['id']

    def get_userid(self):
        """Get userid, username provided in conf"""
        url = '{}/?username={}'.format(
            self.get_url('users'), self.conf['api']['username'])
        res, code = self.get_request(url)

        self.conf['user'] = res[0]

    def get_info(self, kind, uid):
        """Get a object, save to disk and then load it as a JSON object"""

        if self.use_cache:
            # construct cache filename
            filename = self.scratch + '/json/{}.{}.json'.format(kind, uid)
            filename = os.path.expanduser(filename)

            # get from server - no data in cache or empty cache
            if not os.path.exists(filename) or not os.path.getsize(filename):
                # make directory for storing files
                if not os.path.exists(os.path.dirname(filename)):
                    os.makedirs(os.path.dirname(filename))

                url = self.get_url(kind, uid)
                if self.verbose:
                    print >>sys.stderr, 'getting', url
                info, code = self.get_request(url)

                # save to disk
                if info:
                    with open(filename, 'w') as handle:
                        handle.write(json.dumps(info, indent=2))

            # read from cache
            else:
                info = json.loads(open(filename, 'r').read().strip())
                code = 200

        # don't use cache
        else:
            url = self.get_url(kind, uid)
            if self.verbose:
                print >>sys.stderr, 'getting', url
            info, code = self.get_request(url)

        return info, code

    def get_all_info(self, kind):
        """Get all objects of a kind as JSON object"""
        url = self.get_url(kind)
        info, code = self.get_request(url)

        return info

    def get_user(self, uid):
        info, code = self.get_info(kind='users', uid=uid)
        return info

    def get_module(self, uid):
        info, code = self.get_info(kind='modules', uid=uid)
        return info

    def get_workflow(self, uid):
        info, code = self.get_info(kind='workflows', uid=uid)
        return info

    def get_uploads(self):
        """Get all uploads as JSON object"""
        info = self.get_all_info(kind='uploads')
        return info

    def get_upload(self, uid):
        """Get a single upload as JSON object"""
        info, code = self.get_info(kind='uploads', uid=uid)
        return info

    def get_samples(self):
        info = self.get_all_info(kind='samples')
        return info

    def get_sample(self, uid, add_analysis=True):
        info, code = self.get_info(kind='samples', uid=uid)
        if info:
            if add_analysis:
                info = self.add_full_analysis(info)
        return info

    def add_full_analysis(self, sample):
        """Add full analysis info to the sample"""
        analysis_ids = [self.parse_url(uri)['id']
                        for uri in sample['analyses']]
        analyses = [self.get_analysis(uid)[0] for uid in analysis_ids]
        # remove null analyses, probably deleted or no ownership
        analyses = filter(lambda x: x, analyses)

        sample['analyses_full'] = analyses
        return sample

    def get_sample_owner_id(self, sample_id):
        info = self.get_sample(sample_id)
        if info:
            return self.parse_url(info['owner'])['id']
        else:
            return None

    def get_sample_owner(self, sample_id):
        user_id = self.get_sample_owner_id(sample_id)
        if user_id:
            return self.get_user(user_id)
        else:
            return None

    def get_analyses(self):
        info = self.get_all_info(kind='analyses')
        return info

    def get_analysis(self, uid):
        info, code = self.get_info(kind='analyses', uid=uid)
        return info, code

    def get_analysis_owner_id(self, analysis_id):
        info, code = self.get_analysis(analysis_id)
        if info:
            return self.parse_url(info['owner'])['id']
        else:
            return None

    def get_analysis_owner(self, analysis_id):
        """get owner user for analysis"""
        user_id = self.get_analysis_owner_id(analysis_id)
        if user_id:
            return self.get_user(user_id)
        else:
            return None

    def get_group(self, uid):
        info, code = self.get_info(kind='groups', uid=uid)
        return info

    def get_file(self, uid):
        info, code = self.get_info(kind='files', uid=uid)
        return info

    def sample_name_to_id(self, name):
        url = self.get_sample_url()
        info, code = self.get_request(url, user_params={'name': name})

        if not len(info):
            print >>sys.stderr, 'warning: no sample by name', name
            return None
        elif len(info) == 1:
            return info[0]['id']
        else:
            print >>sys.stderr, 'warning: multiple sample by name', name
            return info[0]['id']

    def create_project(self, uid, data):
        url = self.get_project_url()
        res = self.post_request(url, data)

        return res

    def create_sample(self, data, source='api', upload=True):
        """Create sample with the provided info"""
        url = self.get_sample_url()

        if data.get('default_workflow'):
            data['default_workflow'] = '/api/v1/workflows/{}'.format(
                data['default_workflow'])

        sample_id = None
        if 'id' not in data:
            res = self.post_request(url, data)

            # success
            if res.status_code == 201:
                sample_id = self.parse_url(res.headers['location'])['id']
                if self.verbose:
                    print >>sys.stderr, 'created: sample', sample_id

            # failure
            else:
                print >>sys.stderr, 'failed sample creation:', \
                    data['name'], res.status_code, res.reason
        else:
            sample_id = data['id']

        # do the actual upload, update filepath
        files_to_upload = []
        if sample_id:
            order = 0
            for filepath in data.get('filepaths1', []):
                upload_id, filepath, key = self.create_upload(
                    sample_id, filepath, order,
                    is_paired_end=False, source=source)
                files_to_upload.append([upload_id, filepath, key])
                order += 1
            for filepath in data.get('filepaths2', []):
                upload_id, filepath, key = self.create_upload(
                    sample_id, filepath, order,
                    is_paired_end=True, source=source)
                files_to_upload.append([upload_id, filepath, key])
                order += 1

        if upload:
            for upload_id, filepath, key in files_to_upload:
                self.upload_file(upload_id, filepath, key)

        return sample_id

    def create_upload(self, sample_id, filepath, order, is_paired_end,
                      source='api'):
        """Create a upload object and actually upload the file"""
        key = 'uploads/{}/{}/{}'.format(
            self.conf['user']['id'], sample_id, os.path.basename(filepath))
        filesize = os.stat(filepath).st_size if os.path.exists(filepath) else 0

        # update record with location of uploaded file
        url = self.get_upload_url()
        data = {
            'sample': self.get_sample_url(sample_id),
            'key': key,
            'filesize': filesize,
            # 'timetaken': seconds_to_upload,
            'status': 'in_progress',
            'order': order,
            'is_paired_end': is_paired_end,
            'source': source,
        }
        res = self.post_request(url, data)

        if res.status_code == 201:
            upload_id = self.parse_url(res.headers['location'])['id']
            if self.verbose:
                print >>sys.stderr, '\tcreated: upload', upload_id

            return upload_id, filepath, key

        else:
            return None, None, None

    # update file to S3 and update info
    def upload_file(self, upload_id, filepath, key):
        starttime = time.time()
        self.copy_file(filepath, key, action='to')
        seconds_to_upload = int(time.time() - starttime)

        data = {
            'filesize': os.stat(filepath).st_size,
            'status': 'completed',
            'timetaken': seconds_to_upload,
            'seq_length': 0,
        }
        self.update_upload(upload_id, data)

    def create_file(self, uid, data):
        url = self.get_file_url()
        res = self.post_request(url, data)

        return res

    def create_analysis(self, workflow_id, sample_id=None, sample_ids=None,
                        control_id=None, control_ids=None, params=None):
        """Start analysis"""
        analysis_id = None
        url = self.get_analysis_url()
        data = {
            'workflow': '/api/v1/workflows/{}'.format(workflow_id),
            'samples': [],
            'controls': [],
        }

        if sample_id:
            sample_ids = [sample_id]
        if not sample_ids:
            sample_ids = []
        for sample_id in sample_ids:
            data['samples'].append('/api/v1/samples/{}'.format(sample_id))

        if control_id:
            control_ids = [control_id]
        if not control_ids:
            control_ids = []
        for control_id in control_ids:
            data['controls'].append('/api/v1/samples/{}'.format(control_id))

        if params:
            # data['params'] = json.dumps(params)
            data['params'] = params

        res = self.post_request(url, data)

        if res.status_code == 201:
            analysis_id = self.parse_url(res.headers['location'])['id']
            if self.verbose:
                print >>sys.stderr, 'created: analysis {} w/ sample {}'.format(
                    analysis_id, ','.join(sample_ids))

        else:
            print >>sys.stderr, 'failed:', ','.join(sample_ids), \
                res.status_code

        return analysis_id

    def update_upload(self, uid, data):
        url = self.get_upload_url(uid)
        res = self.put_request(url, data)
        return res

    def update_sample(self, uid, data):
        url = self.get_sample_url(uid)
        res = self.put_request(url, data)
        if self.verbose and res.status_code == 204:
            print >>sys.stderr, 'sample', uid, 'updated'
        return res

    def update_project(self, uid, data):
        url = self.get_project_url(uid)
        res = self.put_request(url, data)
        if self.verbose and res.status_code == 204:
            print >>sys.stderr, 'project', uid, 'updated'
        return res

    def update_analysis(self, uid, data):
        url = self.get_analysis_url(uid)
        res = self.put_request(url, data)
        if res.status_code == 204:
            if self.verbose:
                print >>sys.stderr, 'analysis', uid, 'updated'
        else:
            print >>sys.stderr, 'cudnt update analysis {}, code {}'.format(
                uid, res.status_code)
        return res

    def update_file(self, uid, data):
        url = self.get_file_url(uid)
        res = self.put_request(url, data)
        return res

    def delete_sample(self, uid):
        url = self.get_sample_url(uid)
        res = self.delete_request(url)

        if res.status_code == 204:
            if self.verbose:
                print >>sys.stderr, 'deleted sample', uid
        else:
            print >>sys.stderr, 'error: deleting {}, code {}, {}'.format(
                uid, res.status_code, res.reason)
        return res

    def delete_analysis(self, uid):
        url = self.get_analysis_url(uid)
        return self.delete_request(url)

    def copy_file(self, src, dest, action='to'):
        if action == 'to':
            return self.copy_file_to_s3(src, dest)
        else:
            return self.copy_file_from_s3(src, dest)

    def copy_file_from_s3(self, src, dest):
        """Low level function to copy a file from cloud to disk"""
        src = 's3://{}/{}'.format(self.conf['aws']['s3']['bucket'], src)
        cmd = self.get_copy_cmd(src, dest)
        output = subprocess.check_output(
            cmd, stderr=subprocess.STDOUT, shell=True)
        return output

    def copy_file_to_s3(self, src, dest, params=None):
        """Low level function to copy a file to cloud from disk"""
        dest = 's3://{}/{}'.format(self.conf['aws']['s3']['bucket'], dest)

        # check if file is current
        # if self.is_file_current(src, dest):
        #     return False

        cmd = self.get_copy_cmd(src, dest, sse=True, params=params)

        try:
            output = subprocess.check_output(
                cmd, stderr=subprocess.STDOUT, shell=True)
            return output
        except CalledProcessError, e:
            print >>sys.stderr, 'error in copy file to s3'
            print >>sys.stderr, cmd
            print >>sys.stderr, e.output
            return None
        else:
            if self.verbose:
                print >>sys.stderr, 'copying', src, 'to', dest

    def get_copy_cmd(self, src, dest, sse=False, params=None):
        """Create copy cmd based on AWS cli utility"""
        _params = ''
        if sse:
            _params = '--sse'
        if params:
            for arg, val in params.iteritems():
                _params += ' {} "{}"'.format(arg, val)

        cmd = 'aws s3 cp "{}" "{}" {}'.format(src, dest, _params)

        return cmd

    def is_file_current(self, filepath, key_name):
        cmd = 'aws s3 ls {}'.format(key_name)

        statinfo = os.stat(filepath)
        subprocess.check_output(cmd, stderr=subprocess.STDOUT, shell=True)
        key = self.bucket.get_key(key_name)

        if key and statinfo.st_size == key.size:
            return True
        else:
            return False

    def get_history(self, analysis_id):
        """Get SWF history for the analysis"""
        user_id = self.get_analysis_owner_id(analysis_id)
        filekey = 'log/analyses/{}/{}/history.json'.format(
            user_id, analysis_id)
        filename = '{}/history/history.{}.json'.format(
            self.scratch, analysis_id)

        return self.download_file(filekey, filename, load=True, is_json=True)

    def get_filepath(self, filename, dirname=None):
        """Use scratch and [dirname] to construct a full filepath"""

        # move along, nothin to do
        if filename.startswith('/'):
            filepath = filename

        else:
            if dirname:
                filename = '{}/{}'.format(dirname.rstrip('/'), filename)

            # adding dirname made it full
            if filename.startswith('/'):
                filepath = filename
            else:
                filepath = self.scratch + '/' + filename

        return filepath

    def download_file(self, filekey, filename=None, dirname=None, load=False,
                      is_json=False):
        """High level function, downloads to scratch dir
        Open and parse file to JSON if asked
        Uses low level copy_file()
        """
        if not filename:
            if dirname:
                if dirname.startswith('/'):
                    prefix = dirname
                else:
                    prefix = os.path.join(self.scratch, dirname)
            else:
                prefix = self.scratch

            filepath = os.path.join(prefix, os.path.basename(filekey))
        else:
            filepath = self.get_filepath(filename, dirname=dirname)
        filepath = os.path.expanduser(filepath)

        # if file not already there, download it
        if not os.path.exists(filepath) \
                or os.path.getsize(filepath) == 0 \
                or not self.use_cache:
            if self.verbose:
                print >>sys.stderr, 'downloading', filepath

            if not os.path.exists(os.path.dirname(filepath)):
                os.makedirs(os.path.dirname(filepath))

            self.copy_file(filekey, filepath, action='from')
        else:
            if self.verbose:
                print >>sys.stderr, 'exists', filepath

        if load:
            data = open(filepath, 'r').read().strip()
            if is_json:
                data = json.loads(data)
            return data
        else:
            return filepath

    # take a list of files and filter out those that dont have tags
    # if multiple True, return list
    def filter_files_by_tags(self, files, tags, kind=None, exclude=None,
                             multiple=False):
        """Filter files that have all the tags. If exclude tags provided,
        only return the files that don't have the exclude tags.append

        In some cases, multiple files are expected, set it to true.
        """
        if not kind or kind == 'subset':
            files = filter(
                lambda x: x['tags'] and set(tags) <= set(x['tags']), files)
        elif kind == 'exact':
            files = filter(
                lambda x: x['tags'] and set(tags) == set(x['tags']), files)

        if exclude:
            files = filter(
                lambda x: x['tags'] and not set(
                    x['tags']).intersection(set(exclude)),
                files)

        if multiple:
            return files
        else:
            if len(files) == 1:
                return files[0]
            elif len(files) == 0:
                return None
            elif len(files) > 1:
                print >>sys.stderr, 'multiple matches for tags:', tags
                return files[0]

    def get_analysis_files(self, analysis, tags=None, kind='exact',
                           dest=None, dirname=None):
        """For a analysis, go through files and
        get that match the tags
        """
        matching_file = self.filter_files_by_tags(
            analysis['files'], tags, kind=kind)
        filepath = self.download_file(
            matching_file['path'], filename=dest, dirname=dirname)

        return filepath

    # select analyses where one of the files match the tags
    # then get the first matching file
    def get_file_by_tags(self, sample, tags=None,
                         kind=None, exclude=None, analysis_tags=None,
                         dirname=None, multiple=False, dest=None):
        """For a sample, go through analysis and
        get files that match the tags
        """

        # if not dirname:
        #     dirname = ''

        matches = []
        for analysis in sample['analyses_full']:
            if analysis['status'] == 'error':
                continue

            if analysis['params']['info']['genome'] != sample['genome']:
                continue

            if analysis_tags:
                if not analysis['tags']:
                    continue
                if not set(analysis_tags) <= set(analysis['tags']):
                    continue

            matching_file = self.filter_files_by_tags(
                analysis['files'], tags, kind=kind, exclude=exclude,
                multiple=multiple)

            if matching_file:
                if multiple:
                    for f in matching_file:
                        matches.append([analysis['id'], f])
                else:
                    matches.append([analysis['id'], matching_file])

        if len(matches) == 0:
            print >>sys.stderr, 'warning: no matching file for', tags
            return None

        if len(matches) > 1:
            matches.sort(key=lambda x: datetime.datetime.strptime(
                x[1]['last_updated'], '%Y-%m-%dT%H:%M:%S.%f'), reverse=True)
            if not multiple:
                print >>sys.stderr, 'warning: multiple matching file for', tags
            for f in matches:
                print >>sys.stderr, '\t', f[1]['last_updated'], f[1]['path']

        if multiple:
            filepath = []
            for match in matches:
                file1 = match[1]
                path = self.download_file(file1['path'], dirname=dirname)
                filepath.append(path)
        else:
            file1 = matches[0][1]
            # if dest:
            #     dirname = dest
                # filepath = self.download_file(file1['path'], dest)
            # else:
            filepath = self.download_file(
                file1['path'], filename=dest, dirname=dirname)

        return filepath

    # take a list of files and filter out those that dont have node
    # if multiple True, return list
    def filter_files_by_node(self, files, node, multiple=False):
        """Filter files that are from the node.

        In some cases, multiple files are expected, set it to true.
        """
        files = [fl for fl in files if fl['node'] == node]
        # filter(lambda x: x['node'] == node, files)

        if multiple:
            return files
        else:
            if len(files) == 1:
                return files[0]
            elif len(files) == 0:
                return None
            elif len(files) > 1:
                print >>sys.stderr, 'multiple matches for node:', node
                return files[0]

    # select analyses where one of the files match the tags
    # then get the first matching file
    def get_file_by_node(self, sample, node=None, analysis_tags=None,
                         dirname=None, multiple=False, dest=None):
        """For a sample, go through analysis and
        get files that match the tags
        """

        # if not dirname:
        #     dirname = ''

        matches = []
        for analysis in sample['analyses_full']:
            if analysis['status'] == 'error':
                continue

            if analysis_tags:
                if not analysis['tags']:
                    continue
                if not set(analysis_tags) <= set(analysis['tags']):
                    continue

            matching_file = self.filter_files_by_node(
                analysis['files'], node, multiple=multiple)

            if matching_file:
                if multiple:
                    for _ in matching_file:
                        matches.append([analysis['id'], _])
                else:
                    matches.append([analysis['id'], matching_file])

        if len(matches) == 0:
            print >>sys.stderr, 'warning: no matching file for', node
            return None

        if len(matches) > 1:
            matches.sort(key=lambda x: datetime.datetime.strptime(
                x[1]['last_updated'], '%Y-%m-%dT%H:%M:%S.%f'), reverse=True)
            if not multiple:
                print >>sys.stderr, 'warning: multiple matching file for', node
            for _ in matches:
                print >>sys.stderr, '\t', _[1]['last_updated'], _[1]['path']

        if multiple:
            filepath = []
            for match in matches:
                file1 = match[1]
                path = self.download_file(file1['path'], dirname=dirname)
                filepath.append(path)
        else:
            file1 = matches[0][1]
            if dest:
                dirname = dest
                # filepath = self.download_file(file1['path'], dest)
            # else:
            filepath = self.download_file(
                file1['path'], filename=dest, dirname=dirname)

        return filepath

    def get_window_score_filename(self, sample, kind=None, flanking=None):
        suffix = 'score-{}-{}Kb'.format(kind, flanking / 1e3)
        suffix = suffix.replace('.0Kb', 'Kb')

        filename = '{}/score/{}.{}.dedup.{}.txt'.format(
            self.scratch, sample['slug'], sample['genome'], suffix)
        return filename

    def get_expression_count_file(self, sample, multiple=False):
        """Get expression count text file - for RNA-Seq"""
        tags = ['expression_count', 'by_transcript', 'text']
        return self.get_file_by_tags(
            sample, tags, kind='exact', analysis_tags=['alignment'],
            multiple=multiple)

    def get_bigwig_file(self, sample, multiple=False):
        """Get bigwig file - for ChIP-Seq"""
        tags = ['bigwig']
        return self.get_file_by_tags(
            sample, tags, kind='exact', analysis_tags=['alignment'],
            multiple=multiple)

    def get_bam_file(self, sample, tags=None, multiple=False):
        """Get bam file. If you want deduped bam file, call with
        tags = ['bam', 'dedup']
        """
        if not tags:
            tags = ['bam']
        return self.get_file_by_tags(sample, tags, analysis_tags=['alignment'],
                                     multiple=multiple)
